;; ARM code analyser
;; Copyright (C) 2006 Tomasz Malesinski <tmal@mimuw.edu.pl>

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2 of the License, or
;; (at your option) any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software
;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

(load "match.lisp")

(defun single (x)
  (and (consp x) (null (cdr x))))

(defun mapcar-reuse (f list)
  (labels ((rec (l n)
	     (if (null l)
		 list
		 (let ((r (funcall f (car l))))
		   (if (eq r (car l))
		       (rec (cdr l) (1+ n))
		       (append (subseq list 0 n)
			       (cons r (mapcar f (cdr l)))))))))
    (rec list 0)))

(defun image-load-block (filename start end)
  (let ((a (make-array (- end start -1)
		       :element-type '(unsigned-byte 8)))
	(f (open filename :element-type '(unsigned-byte 8))))
    (read-sequence a f)
    (list start end a)))

(defun image-get-byte (image address)
  (if (atom image)
      #xff
      (let ((block (car image)))
	(if (and (>= address (car block))
		 (<= address (cadr block)))
	    (aref (caddr block) (- address (car block)))
	    (image-get-byte (cdr image) address)))))

(defun image-defined-byte (image address)
  (if (atom image)
      nil
      (let ((block (car image)))
	(if (and (>= address (car block))
		 (<= address (cadr block)))
	    t
	    (image-defined-byte (cdr image) address)))))

(defun image-defined-word (image address)
  (and
   (image-defined-byte image address)
   (image-defined-byte image (+ address 1))
   (image-defined-byte image (+ address 2))
   (image-defined-byte image (+ address 3))))

(defun image-get-word (image address)
  (+ (image-get-byte image address)
     (* (image-get-byte image (1+ address)) #x100)
     (* (image-get-byte image (+ 2 address)) #x10000)
     (* (image-get-byte image (+ 3 address)) #x1000000)))

(defun image-get-sequence (image address n)
  (if (<= n 0)
      nil
      (cons (image-get-byte image address)
	    (image-get-sequence image (1+ address) (1- n)))))

(defun bitmatch (a b mask)
  (= (logand a mask) (logand b mask)))

(defun getbits (n a b)
  (ldb (byte b a) n))

(defun bitset (n a)
  (= (ldb (byte 1 a) n) 1))

(defun ror (n count)
  (logand (logior (ash n (- count)) (ash n (- 32 count))) #xffffffff))

(defun sign-extend (n bits)
  (if (bitset n (1- bits))
      (- n (ash 1 bits))
      n))

(defun power2p (n)
  (labels ((rec (n a)
	     (cond
	       ((= n 1) a)
	       ((oddp n) nil)
	       (t (rec (/ n 2) (1+ a))))))
    (if (/= n 0) (rec n 0))))

(defconstant arm-condition-vector
  (vector 'eq 'ne 'cs 'cc 'mi 'pl 'vs 'vc 'hi 'ls 'ge 'lt 'gt 'le 'al 'nv))

(defun arm-condition (code)
  (aref arm-condition-vector code))

(defun arm-reg (n)
  (list 'reg n))

(defun arm-regd (opcode)
  (arm-reg (getbits opcode 12 4)))

(defun arm-regn (opcode)
  (arm-reg (getbits opcode 16 4)))

(defun arm-regs (opcode)
  (arm-reg (getbits opcode 8 4)))

(defun arm-regm (opcode)
  (arm-reg (getbits opcode 0 4)))

(defun arm-bits (opcode)
  (if (bitset opcode 20) 's 'ns))

(defun arm-shifter-operand (opcode)
  (cond
    ((bitmatch opcode #x02000000 #x0e000000)
     (ror (getbits opcode 0 8) (* (getbits opcode 8 4) 2)))
    ((bitmatch opcode #x00000000 #x0e000010)
     (arm-shift-imm opcode))
    ((bitmatch opcode #x00000010 #x0e000090)
     (let ((rm (arm-reg (getbits opcode 0 4)))
	   (rs (arm-reg (getbits opcode 8 4)))
	   (shift (getbits opcode 5 2)))
       (list
	(cond ((= shift 0) 'lsl)
	      ((= shift 1) 'lsr)
	      ((= shift 2) 'asr)
	      ((= shift 3) 'ror))
	rm rs)))
    (t 'shifter)))

(defun arm-data-processing-mnemonic (op)
  (aref (vector 'and 'eor 'sub 'rsb 'add 'adc 'sbc 'rsc
		'tst 'teq 'cmp 'cmn 'orr 'mov 'bic 'mvn)
	op))
	  
(defun arm-disassemble-data-processing (opcode)
  (let* ((op (getbits opcode 21 4))
	 (m (arm-data-processing-mnemonic op)))
    (cond
      ((member m '(mov mvn))
       (list m (arm-regd opcode) (arm-shifter-operand opcode) (arm-bits opcode)))
      ((member m '(cmp cmn tst teq))
       (list m (arm-regn opcode) (arm-shifter-operand opcode)))
      (t
       (list m (arm-regd opcode) (arm-regn opcode) (arm-shifter-operand opcode)
	     (arm-bits opcode))))))

(defun arm-shift-imm (opcode)
  (let ((rm (arm-reg (getbits opcode 0 4)))
	(imm (getbits opcode 7 5))
	(shift (getbits opcode 5 2)))
    (if (= shift 3)
	(if (= imm 0)
	    (list 'rrx rm)
	    (list 'ror rm imm))
	(list
	 (cond ((= shift 0) 'lsl)
	       ((= shift 1) 'lsr)
	       ((= shift 2) 'asr))
	 rm imm))))


(defun arm-addressing-mode-2 (opcode)
  (let ((offset
	 (cond ((bitmatch opcode #x04000000 #x0e000000)
		(getbits opcode 0 12))
	       ((bitmatch opcode #x06000000 #x0e000ff0)
		(arm-reg (getbits opcode 0 4)))
	       ((bitmatch opcode #x06000000 #x0e000010)
		(arm-shift-imm opcode))
	       (t 'unknown))))
    (list 
     (cond ((bitmatch opcode #x00000000 #x01200000)
	    'post)
	   ((bitmatch opcode #x01000000 #x01200000)
	    'mem)
	   ((bitmatch opcode #x01200000 #x01200000)
	    'pre)
	   (t 't))
     (list
      (if (bitset opcode 23) '+ '-)
      (arm-regn opcode)
      offset))))

(defun arm-disassemble-load-store (opcode)
  (cond
    ((bitmatch opcode #x04100000 #x0c100000)
     (list 'ldr (arm-regd opcode) (arm-addressing-mode-2 opcode)
	   (if (= (getbits opcode 22 1) 0) 'w 'b)))
    ((bitmatch opcode #x04000000 #x0c100000)
     (list 'str (arm-regd opcode) (arm-addressing-mode-2 opcode)
	   (if (= (getbits opcode 22 1) 0) 'w 'b)))
    (t 'load-store)))
  
(defun arm-disassemble-load-store-half (opcode)
  (list
   (if (bitset opcode 20) 'ldr 'str)
   (arm-regd opcode)
   (list
    (cond ((bitmatch opcode #x00000000 #x01200000)
	   'post)
	  ((bitmatch opcode #x01000000 #x01200000)
	   'mem)
	  ((bitmatch opcode #x01200000 #x01200000)
	   'pre)
	  (t 't))
    (list
     (if (bitset opcode 23) '+ '-)
     (arm-regn opcode)
     (if (bitset opcode 22)
	 (+ (* (getbits opcode 8 4) 16) (getbits opcode 0 4))
	 (arm-regm opcode))))
   (case (getbits opcode 5 2)
     (#b01 'h)
     (#b10 'sb)
     (#b11 'sh))))

(defun arm-branch-offset (opcode)
  (* 4 (sign-extend (getbits opcode 0 24) 24)))

(defun arm-disassemble-load-store-multiple (opcode)
  (list
   (if (bitset opcode 20) 'ldm 'stm)
   (arm-regn opcode)
   (do ((i 15 (1- i)) regl)
       ((< i 0) regl)
     (if (bitset opcode i) (push (arm-reg i) regl)))
   (if (bitset opcode 23)
       (if (bitset opcode 24) 'ib 'ia)
       (if (bitset opcode 24) 'db ' da))
   (if (bitset opcode 21) 'w 'nw)
   (if (bitset opcode 22) 's 'ns)))
	   
(defun arm-disassemble-arithm-ext (opcode)
  (let ((regn (arm-reg (getbits opcode 12 4)))
	(regs (arm-regs opcode))
	(regd (arm-reg (getbits opcode 16 4)))
	(regm (arm-regm opcode))
	(s (if (bitset opcode 20) 's 'ns)))
    (case (getbits opcode 21 3)
      (#b000
       (list 'mul regd regm regs s))
      (#b001
       (list 'mla regd regm regs regn s))
      (#b111
       (list 'smlal regn regd regm regs s))
      (#b110
       (list 'smlal regn regd regm regs s))
      (#b101
       (list 'umlal regn regd regm regs s))
      (#b100
       (list 'umull regn regd regm regs s)))))

(defun arm-disassemble-opcode (opcode)
  (cons (arm-condition (ldb (byte 4 28) opcode))
	(cond
	  ((bitmatch opcode #x00000090 #x0f0000f0)
	   (arm-disassemble-arithm-ext opcode))
	  ((bitmatch opcode #x00000090 #x0e000090)
	   (arm-disassemble-load-store-half opcode))
	  ((bitmatch opcode 0 #x0c000000)
	   (arm-disassemble-data-processing opcode))
	  ((bitmatch opcode #x04000000 #x0c000000)
	   (arm-disassemble-load-store opcode))
	  ((bitmatch opcode #x0a000000 #x0f000000)
	   (list 'b (arm-branch-offset opcode)))
	  ((bitmatch opcode #x0b000000 #x0f000000)
	   (list 'bl (arm-branch-offset opcode)))
	  ((bitmatch opcode #x08000000 #x0e000000)
	   (arm-disassemble-load-store-multiple opcode))
	  (t '(unknown)))))

(defun arm-disassemble-instruction (code)
  (if (< (length code) 4)
      (cons '(al unknown) (length code))
      (let ((opcode
	     (+ (car code) (* (cadr code) #x100) (* (caddr code) #x10000)
		(* (cadddr code) #x1000000))))
	(cons (arm-disassemble-opcode opcode) 4))))

(defun arm-disassemble-range (image start end)
  (if (> start end)
      nil
      (let ((i (arm-disassemble-instruction
		(image-get-sequence image start 4))))
	(cons (car i)
	      (arm-disassemble-range image
				     (+ start (cdr i)) end)))))

(defun sem-size-to-function (size)
  (case size
    (w 'mem)
    (b 'memb)
    (sb 'memsb)
    (h 'memh)
    (sh 'memsh)))

(defun arm-sem-mem-operand (operand)
  (let ((f (sem-size-to-function (second operand))))
    (cond-matchg (car operand)
		 ((('mem ?a))
		  (cons (list f ?a) nil))
		 ((('post (?s ?r ?o)))
		  (cons (list f ?r)
			(list (cons ?r (cadar operand)))))
		 ((('pre (?s ?r ?o)))
		  (cons (list f (cadar operand))
			(list (cons ?r (cadar operand))))))))

(defun arm-sem-set-pc-to-next ()
  (list 'set (list 'reg 15)
	(list '- (list 'reg 15) 4)))

(defun arm-sem-sets-pc (sem)
  (find-if #'(lambda (x)
	       (and (eq (car x) 'set)
		    (equal (cadr x) '(reg 15))))
	   sem))

(defun arm-sem-prepend-condition (condition sem)
  (cond
    ((null condition) (list (arm-sem-set-pc-to-next)))
    ((eq condition 't)
     (if (arm-sem-sets-pc sem)
	 sem
	 (cons (arm-sem-set-pc-to-next) sem)))
    (t (let ((sem1
	      (mapcar #'(lambda (x)
			  (if-match ('set ('reg 15) ?v) x
				    (list 'set (list 'reg 15)
					  (list 'if condition
						?v
						(list '- (list 'reg 15) 4)))
				    (list 'if-then condition x)))
		      sem)))
	 (if (arm-sem-sets-pc sem)
	     sem1
	     (cons (arm-sem-set-pc-to-next) sem1))))))

(defun arm-sem-condition-and-pc (condition sem)
  (arm-sem-prepend-condition
   (cdr (assoc condition
	       '((al . t)
		 (eq . (reg z))
		 (ne . (= (reg z) 0))
		 (cs . (reg c))
		 (cc . (= (reg c) 0))
		 (mi . (reg n))
		 (pl . (= (reg n) 0))
		 (vs . (reg v))
		 (vc . (= (reg v) 0))
		 (hi . (and (reg c) (= (reg z) 0)))
		 (ls . (or (= (reg c) 0) (reg z)))
		 (ge . (= (reg n) (reg v)))
		 (lt . (/= (reg n) (reg v)))
		 (gt . (and (= (reg n) (reg v)) (= (reg z) 0)))
		 (le . (or (/= (reg n) (reg v)) (reg z)))
		 (nv . nil))))
   sem))

(defun arm-sem-shifter (operand)
  '(((reg c) . unknown)))

(defun arm-sem-set-nz (value)
  (list
   (cons '(reg n) (list '< value 0))
   (cons '(reg z) (list '= value 0))))

(defun arm-sem-instruction-bit (pattern args)
  (let ((res (subst (second args) '?x
		    (subst (third args) '?y pattern))))
    (cons
     (cons (first args) res)
     (if (eq (fourth args) 's)
	 (append
	  (arm-sem-set-nz res)
	  (arm-sem-shifter (third args)))))))

(defun arm-sem-instruction-bit-test (pattern args)
  (let ((res (subst (first args) '?x
		    (subst (second args) '?y pattern))))
    (append
     (arm-sem-set-nz res)
     (arm-sem-shifter (second args)))))

(defun arm-sem-instruction-arithm (op args extra-terms carry-sym overflow-sym)
  (let ((res (append
	      (list op (second args) (third args))
	      extra-terms)))
    (cons
     (cons (first args) res)
     (if (eq (fourth args) 's)
	 (append
	  (arm-sem-set-nz res)
	  (list
	   (cons '(reg c) (list carry-sym
				(second args) (third args)))
	   (cons '(reg v) (list overflow-sym
				(second args) (third args)))))))))

(defun arm-sem-instruction-cmp (op args carry-sym overflow-sym)
  (let ((res (list op (first args) (second args))))
    (append
     (arm-sem-set-nz res)
     (list
      (cons '(reg c) (list carry-sym
			   (first args) (second args)))
      (cons '(reg v) (list overflow-sym
			   (first args) (second args)))))))

(defun arm-sem-load-store-multiply (args load)
  (let ((regn (first args))
	(regl (second args))
	(mode (third args))
	(write (fourth args)))
    (let ((p
	   (cond
	     ((eq mode 'ia)
	      (cons 0 (* (length regl) 4)))
	     ((eq mode 'ib)
	      (cons 4 (* (length regl) 4)))
	     ((eq mode 'da)
	      (cons (- 4 (* (length regl) 4)) (- (* (length regl) 4))))
	     ((eq mode 'db)
	      (cons (- (* (length regl) 4)) (- (* (length regl) 4)))))))
      (do ((r regl (cdr r))
	   (offset (car p) (+ offset 4))
	   (s (if (or (member regn regl) (eq write 'nw))
		  nil
		  (list (cons regn (list '+ regn (cdr p)))))
	      (cons 
	       (let ((mem (list 'mem (list '+ regn offset))))
		 (if load
		     (cons (car r) mem)
		     (cons mem (car r))))
	       s)))
	  ((not r) s)))))

(defun set-statements (pairs)
  (mapcar #'(lambda (x) (list 'set (car x) (cdr x)))
	  pairs))

(defun arm-sem-instruction (instruction)
  (arm-sem-condition-and-pc
   (car instruction)
   (set-statements
    (let ((instr (cadr instruction))
	  (args (cddr instruction)))
      (cond ((eq instr 'mov)
	     (cons
	      (cons (first args) (second args))
	      (if (eq (third args) 's)
		  (append
		   (arm-sem-set-nz (second args))
		   (arm-sem-shifter (second args))))))

	    ((eq instr 'mvn)
	     (let ((res (list 'lognot (second args))))
	       (cons
		(cons (first args) res)
		(if (eq (third args) 's)
		    (append
		     (arm-sem-set-nz res)
		     (arm-sem-shifter (second args)))))))

	    ((eq instr 'add)
	     (arm-sem-instruction-arithm
	      '+ args nil
	      'carry-add 'overflow-add))

	    ((eq instr 'adc)
	     (arm-sem-instruction-arithm
	      '+ args '((reg c))
	      'carry-adc 'overflow-adc))

	    ((eq instr 'sub)
	     (arm-sem-instruction-arithm
	      '- args nil
	      '>=u 'overflow-sub))

	    ((eq instr 'sbc)
	     (arm-sem-instruction-arithm
	      '- args '((lognot (reg c)))
	      'carry-sbc 'overflow-sbc))

	    ((eq instr 'rsb)
	     (arm-sem-instruction-arithm
	      '-
	      (cons (first args)
		    (cons (third args)
			  (cons (second args)
				(cdddr args))))
	      nil
	      '>=u 'overflow-sub))

	    ((eq instr 'rsc)
	     (arm-sem-instruction-arithm
	      '-
	      (cons (first args)
		    (cons (third args)
			  (cons (second args)
				(cdddr args))))
	      '((lognot (reg c)))
	      'carry-sbc 'overflow-sbc))
	   
	    ((eq instr 'cmp)
	     (arm-sem-instruction-cmp '- args '>=u 'overflow-sub))

	    ((eq instr 'cmn)
	     (arm-sem-instruction-cmp '+ args 'carry-add 'overflow-add))

	    ((eq instr 'and)
	     (arm-sem-instruction-bit '(logand ?x ?y) args))

	    ((eq instr 'eor)
	     (arm-sem-instruction-bit '(logxor ?x ?y) args))

	    ((eq instr 'orr)
	     (arm-sem-instruction-bit '(logior ?x ?y) args))

	    ((eq instr 'bic)
	     (arm-sem-instruction-bit '(logand ?x (lognot ?y)) args))

	    ((eq instr 'tst)
	     (arm-sem-instruction-bit-test '(logand ?x ?y) args))

	    ((eq instr 'teq)
	     (arm-sem-instruction-bit-test '(logxor ?x ?y) args))

	    ((eq instr 'mul)
	     (let ((res (list '* (second args) (third args))))
	       (cons
		(cons (first args) res)
		(if (eq (fourth args) 's)
		    (arm-sem-set-nz res)
		    ()))))

	    ((eq instr 'mla)
	     (let ((res (list '+ (list '* (second args) (third args))
			      (fourth args))))
	       (cons
		(cons (first args) res)
		(if (eq (fifth args) 's)
		    (arm-sem-set-nz res)
		    ()))))
	     
	    ((member instr '(smlal smull umlal umull))
	     (cons (cons (first args) 'unknown)
		   (cons (cons (second args) 'unknown)
			 (if (eq (fifth args) 's)
			     (arm-sem-set-nz 'unknown)
			     ()))))
	     
	    ((eq instr 'b)
	     (list (cons '(reg 15)
			 (list '+ '(reg 15) (first args)))))

	    ((eq instr 'bl)
	     (list 
	      (cons '(reg 14) (list '- '(reg 15) 4))
	      (cons '(reg 15) (list '+ '(reg 15) (first args)))))

	    ((eq instr 'ldr)
	     (let ((p (arm-sem-mem-operand (cdr args))))
	       (cons (cons (first args)(car p))
		     (cdr p))))

	    ((eq instr 'str)
	     (let ((p (arm-sem-mem-operand (cdr args))))
	       (cons (cons (car p) (first args))
		     (cdr p))))

	    ((eq instr 'ldm)
	     (arm-sem-load-store-multiply args t))

	    ((eq instr 'stm)
	     (arm-sem-load-store-multiply args nil))

	    (t nil))))))

(defun arm-mnemonic-simple (instr &optional s)
  (strcat
   (symbol-name (second instr))
   (case (first instr)
     (al "")
     (t (symbol-name (first instr))))
   (if (eq s 's) "S" "")))

(defun arm-reg-or-imm-to-string (a)
  (cond
    ((numberp a)
     (strcat "#" (expr-to-string a)))
    (t (expr-to-string a))))

(defun arm-shifter-to-string (shift)
  (cond-matchg
   shift
   (((?sh ?x ?y) (member ?sh '(lsl lsr asr ror)))
    (if (eql 0 ?y)
	(arm-reg-or-imm-to-string ?x)
	(strcat (arm-reg-or-imm-to-string ?x) ", " (symbol-name ?sh) " "
		(arm-reg-or-imm-to-string ?y))))

   ((('rrx ?r))
    (strcat (arm-reg-or-imm-to-string ?r) ", RRX"))
    
   ((?x) (arm-reg-or-imm-to-string ?x))))

(defun arm-addr-mode2-to-string (mode)
  (destructuring-bind (ind (sign regn offset)) mode
    (let ((sregn (arm-reg-or-imm-to-string regn))
	  (soffset
	   (cond
	     ((eql offset 0) "")
	     ((and (numberp offset) (eq sign '-))
	      (strcat ", #-" (expr-to-string offset)))
	     (t (strcat ", "
			(if (eq sign '+) "" "-")
			(arm-shifter-to-string offset))))))
      (case ind
	(mem
	 (strcat "[" sregn soffset "]"))
	(post
	 (strcat "[" sregn "]" soffset))
	(pre
	 (strcat "[" sregn soffset "]!"))))))

(defun arm-reg-set-to-string (set)
  (labels ((rec (set acc first)
	     (if (null set)
		 acc
		 (let ((r (arm-reg-or-imm-to-string (car set))))
		   (rec (cdr set)
			(if first r (strcat acc ", " r))
			nil)))))
    (strcat "{" (rec set "" t) "}")))

(defun arm-instruction-to-string (instr addr)
  (let ((p
	 (case (second instr)
	   ((b bl)
	    (cons (arm-mnemonic-simple instr)
		  (expr-to-string (+ addr (third instr) 8))))
	   
	   ((mov mvn cmp cmn tst teq)
	    (cons
	     (arm-mnemonic-simple instr (fifth instr))
	     (strcat (expr-to-string (third instr)) ", "
		     (arm-shifter-to-string (fourth instr)))))

	   ((add sub rsb adc sbc rsc and bic eor orr)
	    (cons
	     (arm-mnemonic-simple instr (sixth instr))
	     (strcat (expr-to-string (third instr)) ", "
		     (expr-to-string (fourth instr)) ", "
		     (arm-shifter-to-string (fifth instr)))))

	   ((ldr str)
	    (cons
	     (let ((m (arm-mnemonic-simple instr)))
	       (if (eq (fifth instr) 'w)
		   m
		   (strcat m (symbol-name (fifth instr)))))
	     (strcat (arm-reg-or-imm-to-string (third instr)) ", "
		     (arm-addr-mode2-to-string (fourth instr)))))

	   ((ldm stm)
	    (cons
	     (strcat
	      (arm-mnemonic-simple instr)
	      (symbol-name (fifth instr)))
	     (strcat (arm-reg-or-imm-to-string (third instr))
		     (if (eq (sixth instr) 'w) "!" "") ", "
		     (arm-reg-set-to-string (fourth instr))
		     (if (eq (seventh instr) 's) "^" ""))))

	   ((mul)
	    (cons
	     (arm-mnemonic-simple instr (sixth instr))
	     (strcat (arm-reg-or-imm-to-string (third instr)) ", "
		     (arm-reg-or-imm-to-string (fourth instr)) ", "
		     (arm-reg-or-imm-to-string (fifth instr)))))

	   ((mla smlal smull umlal umull)
	    (cons
	     (arm-mnemonic-simple instr (seventh instr))
	     (strcat (arm-reg-or-imm-to-string (third instr)) ", "
		     (arm-reg-or-imm-to-string (fourth instr)) ", "
		     (arm-reg-or-imm-to-string (fifth instr)) ", "
		     (arm-reg-or-imm-to-string (sixth instr)))))

	   (t (cons "???" (format nil "~a" instr))))))
    (format nil "~8a ~a" (car p) (cdr p))))

(defun reg< (reg1 reg2)
  (let ((n1 (second reg1))
	(n2 (second reg2)))
    (if (numberp n1)
	(if (numberp n2)
	    (< n1 n2)
	    t)
	(if (numberp n2)
	    nil
	    (string< n1 n2)))))

(defun expr< (e1 e2)
  (cond
    ((numberp e1)
     (if (numberp e2) (< e1 e2) t))
    ((symbolp e1)
     (cond
       ((numberp e2) nil)
       ((symbolp e2) (string< e1 e2))
       (t t)))
    ((consp e1)
     (if (consp e2)
	 (if (equal (car e1) (car e2))
	     (expr< (cdr e1) (cdr e2))
	     (expr< (car e1) (car e2)))
	 nil))))

(defun split-list (pred list)
  (let (lt lf)
    (dolist (x list)
      (if (funcall pred x)
	  (push x lt)
	  (push x lf)))
    (cons (reverse lt) (reverse lf))))

(defun expr-substitute (expr vars &optional (atom? #'atom))
  (let ((s (assoc expr vars :test #'equal)))
    (cond
      ((funcall atom? expr) expr)
      (s (cdr s))
      ((consp expr)
       (mapcar-reuse #'(lambda (x) (expr-substitute x vars atom?))
		     expr)))))

(defun expr-subst-simplify (expr vars &key (atom? #'atom) max-depth)
  (let ((s (assoc expr vars :test #'equal)))
    (cond
      ((and max-depth (< max-depth 0)) expr)
      ((funcall atom? expr) expr)
      (s (cdr s))
      ((consp expr)
       (let ((new
	      (let ((new-max-depth (if max-depth (1- max-depth) nil)))
		(mapcar-reuse
		 #'(lambda (x)
		     (expr-subst-simplify x vars
					  :atom? atom?
					  :max-depth new-max-depth))
		 expr))))
	 (if (eq expr new)
	     expr
	     (expr-simplify-1 new)))))))

(defmacro aexpr-rec (expr &body body)
  `(labels ((rec (it)
	     (labels ((rec-all ()
			(if (consp it)
			    (mapcar-reuse #'rec it)
			    it)))
	       ,@body)))
    (rec ,expr)))
		    
(defun expr-is-condition (expr)
  (or (eql expr 0)
      (eql expr 1)
      (and (consp expr)
	   (member (car expr)
		   '(and or not /= = < > <= >=
		     <u >u <=u >=u)))
      (if-match ('if _ ?t ?e) expr
		(and (expr-is-condition ?t)
		     (expr-is-condition ?e))
		nil)))

(defun negate-condition (cond)
  (cond-matchg cond
	       ((('= ?x ?y))
		(list '/= ?x ?y))
	       ((('/= ?x ?y))
		(list '= ?x ?y))
	       ((('< ?x ?y))
		(list '>= ?x ?y))
	       ((('> ?x ?y))
		(list '<= ?x ?y))
	       ((('<= ?x ?y))
		(list '> ?x ?y))
	       ((('>= ?x ?y))
		(list '< ?x ?y))
	       ((('>=u ?x ?y))
		(list '<u ?x ?y))
	       ((('and &rest ?a))
		(cons 'or (mapcar #'negate-condition ?a)))
	       ((('or &rest ?a))
		(cons 'and (mapcar #'negate-condition ?a)))
	       ((('not ?x))
		?x)
	       ((0) 1)
	       ((1) 0)
	       ((?x) (list 'not ?x))))

(defun find-= (val true)
  (some #'(lambda (x)
	    (if-match ('= val ?n) x
		      (if (numberp ?n) ?n nil)
		      nil))
	true))
    
(defun test-condition (true cond)
  (cond-matchg
   cond
    ;; TODO - dirty hack in order not to simplify (reg ..) in vrefs
   ((('reg _))
    nil)
   (((?cmp ?x ?n)
     (and (numberp ?n)
	  (member ?cmp '(< > <= >= <u >u <=u >=u))
	  (find-= ?x true)))
    (let ((?v (find-= ?x true)))
      (case ?cmp
	;; TODO - correct signed/unsigned comparison
	((< <u) (if (< ?v ?n) 't 'f))
	((> >u) (if (> ?v ?n) 't 'f))
	((<= <=u) (if (<= ?v ?n) 't 'f))
	((>= >=u) (if (>= ?v ?n) 't 'f)))))
   ((_)
    (cond
      ((member cond true :test #'equal)
       't)
      ((member (negate-condition cond) true :test #'equal)
       'f)
      (t nil)))))

(defun expr-simplify-with-true (true expr)
  (aexpr-rec expr
	     ;; TODO - true should be extended with ?c/~?c
	     (cond-matchg
	      it
	      ((('if ?c ?t ?e))
	       (case (test-condition true ?c)
		 ((t) (rec ?t))
		 ((f) (rec ?e))
		 ((nil) (rec-all))))
	      ((_)
	       (case (test-condition true it)
		 ((t) 1)
		 ((f) 0)
		 ((nil) 
		  (let ((n (rec-all)))
		    (if (not (eq n it))
			(expr-simplify-1 n)
			n))))))))
			    

(defun expr-simplify-cond (expr)
  (cond-match
   expr
   (('if ?c ?t ?e)
    (expr-simplify-1
     (list 'or
	   (expr-simplify-1 (list 'and ?c ?t))
	   (expr-simplify-1 (list 'and (negate-condition ?c)
				  ?e)))))
   (_ expr)))

(defun expr-simplify-const (f expr)
  (if (find-if-not #'numberp (cdr expr))
      expr
      (apply f (cdr expr))))

(defun expr-simplify-assoc-com (op args neutral)
  (labels ((rec (args cl ol)
	     (if (null args)
		 (list cl ol)
		 (cond
		   ((numberp (car args))
		    (rec (cdr args) (cons (car args) cl) ol))
		   ((consp (car args))
		    (if (eq (caar args) op)
			(let ((p (rec (cdar args) cl ol)))
			  (rec (cdr args) (first p) (second p)))
			(rec (cdr args) cl (cons (car args) ol))))
		   (t (rec (cdr args) cl (cons (car args) ol)))))))
    (destructuring-bind (cl ol) (rec args nil nil)
      (let* ((c (apply (symbol-function op) cl))
	     (r (if (= c neutral)
		    (nreverse ol)
		    (cons c (nreverse ol)))))
	(cond ((null r) neutral)
	      ((single r) (car r))
	      (t (cons op r)))))))

(defun expr-negate (expr)
  (cond-matchg
   expr
   ((?x (numberp ?x)) (- ?x))
   ((('+ &rest ?args))
    (cons '+ (mapcar #'(lambda (x) (expr-negate x)) ?args)))
   ((('* ?n &rest ?args) (numberp ?n))
    (nconc (list '* (- ?n)) ?args))
   ((('* &rest ?args))
    (nconc (list '* -1) ?args))
   ((('- ?x)) ?x)
   ((?x)
    (list '- ?x))))

(defun expr-simplify-pairwise (op args f)
  (labels ((rec (args res)
	     (cond
	       ((null args)
		res)
	       ((single args)
		(rec nil (cons (car args) res)))
	       (t
		(let ((r (funcall f (car args)
				  (cadr args))))
		  (if r
		      (rec (cons r (cddr args)) res)
		      (rec (cdr args)
			   (cons (car args) res))))))))
    (let ((res (rec args nil)))
      (if (single res)
	  (car res)
	  (cons op (nreverse res))))))

(defun expr-simplify-plus-distr (expr)
  (labels ((comp (p1 p2)
	     (let ((c (+ (car p1) (car p2))))
	       (case c
		 (1 (cdr p1))
		 (-1 (list '- (cdr p1)))
		 (t (list '* c (cdr p1))))))
	   (decomp (expr)
	     (cond-matchg
	      expr
	      ((('* ?n ?a) (numberp ?n))
	       (cons ?n ?a))
	      ((('- ?a))
	       (cons -1 ?a))
	      ((?a)
	       (cons 1 ?a))))
	   (distr2 (a b)
	     (let ((p1 (decomp a))
		   (p2 (decomp b)))
	       (if (or (null p1)
		       (null p2)
		       (not (equal (cdr p1)
				   (cdr p2))))
		   nil
		   (comp p1 p2)))))
    (if (and (consp expr) (eq '+ (car expr)))
	(expr-simplify-pairwise '+ (cdr expr) #'distr2)
	expr)))

(defun expr-cmp-op-without-eq-p (op)
  (member op '(< > <u >u)))

(defun expr-cmp-op-with-eq-p (op)
  (member op '(<= >= <=u >=u)))

(defun expr-cmp-op-eq-p (op)
  (member op '(= /=)))

(defun expr-cmp-op-with-eq (op)
  (case op (< '<=) (> '>=) (<u '<=u) (>u '>=u)))

(defun expr-cmp-op-without-eq (op)
  (case op (<= '<) (>= '>) (<=u '<u) (>=u '>u)))

(defun expr-reverse-cmp-op (op)
  (cdr (assoc op '((= . =) (/= . /=)
		   (< . >) (> . <) (<= . >=) (>= . <=)
		   (<u . >u) (>u . <u) (<=u . >=u) (>=u . <=u)))))

(defun expr-simplify-and-or-cmp (expr)
  (cond-matchg
   expr
   ((('or (?cmp ?a ?b) ('= ?c ?d))
     (and (expr-cmp-op-without-eq-p ?cmp)
	  (let ((ab (expr-simplify-1 (list '- ?a ?b))))
	    (or (equal ab (expr-simplify-1 (list '- ?c ?d)))
		(equal ab (expr-simplify-1 (list '- ?d ?c)))))))
    (list (expr-cmp-op-with-eq ?cmp) ?a ?b))
   ((('and (?cmp ?a ?b) ('/= ?c ?d))
     (and (expr-cmp-op-with-eq-p ?cmp)
	  (let ((ab (expr-simplify-1 (list '- ?a ?b))))
	    (or (equal ab (expr-simplify-1 (list '- ?c ?d)))
		(equal ab (expr-simplify-1 (list '- ?d ?c)))))))
    (list (expr-cmp-op-without-eq ?cmp) ?a ?b))
   ((('and (?ge ?a ?n1) (?ge ?b ?n2))
     (and (member ?ge '(>= >=u))
	  (numberp ?n2)
	  (> ?n2 0)
	  (equal ?b (expr-simplify-1 (list '- ?a ?n1)))))
    (list ?ge ?a (expr-simplify-1 (list '+ ?n1 ?n2))))
   ((('or (?lt ?a ?n1) (?lt ?b ?n2))
     (and (member ?lt '(< <u))
	  (numberp ?n1)
	  (> ?n1 0)
	  (equal ?a (expr-simplify-1 (list '- ?b ?n2)))))
    (list ?lt ?b (expr-simplify-1 (list '+ ?n1 ?n2))))
   ((_) expr)))
   
(defun expr-simplify-and-or (args and-mode)
  (labels ((rec (left right)
	     (if (null right)
		 (nreverse left)
		 (let ((true (if and-mode
				 (append left (cdr right))
				 (mapcar #'negate-condition
					 (append left (cdr right))))))
		   (rec (cons (expr-simplify-with-true true
						       (car right))
			      left)
			(cdr right)))))
	   (filter-const (args res)
	     (if (null args)
		 (nreverse res)
		 (cond
		   ((eql (car args) (if and-mode 0 1))
		    (list (car args)))
		   ((eql (car args) (if and-mode 1 0))
		    (filter-const (cdr args) res))
		   (t (filter-const (cdr args)
				    (cons (car args) res)))))))
    (let ((new-args 
	   (filter-const
	    (rec ()
		 (mapcar #'expr-simplify-cond args))
	    ())))
      (cond
	((null new-args) (if and-mode 1 0))
	((single new-args) (car new-args))
	(t
	 (expr-simplify-and-or-cmp
	  (cons (if and-mode 'and 'or) new-args)))))))

(defun expr-simplify-cmp-const-to-right (expr)
  (if (and (numberp (second expr))
	   (not (numberp (third expr))))
      (list (expr-reverse-cmp-op (first expr))
	    (third expr) (second expr))
      expr))

(defun expr-simplify-eq-consts-to-right (expr)
  (if-match (?cmp ('+ ?n1 &rest ?x) ?n2) expr
	    (if (and (member ?cmp '(= /=))
		     (numberp ?n1)
		     (numberp ?n2)
		     ?x)
		(list ?cmp (expr-simplify-1 (cons '+ ?x))
		      (- ?n2 ?n1))
		expr)
	    expr))

(defun expr-simplify-eq-0 (expr)
  (if-match (?cmp ('+ ?x ?y) 0) expr
	    (if (and (member ?cmp '(= /=)))
		(list ?cmp ?x (expr-negate ?y))
		expr)
	    expr))

(defun expr-negative-simpler (expr)
  (or (match ('- _) expr)
      (match ('+ ('- _) &rest _) expr)))

(defun expr-simplify-eq-minus (expr)
  (let ((op (first expr))
	(left (second expr))
	(right (third expr)))
    (if (and (member op '(= /=))
	     (or (numberp right)
		 (expr-negative-simpler right))
	     (expr-negative-simpler left))
	(list op (expr-negate left) (expr-negate right))
	expr)))

(defun expr-simplify-cmp (expr)
  (expr-simplify-eq-minus
   (expr-simplify-eq-0
    (expr-simplify-eq-consts-to-right
     (expr-simplify-cmp-const-to-right expr)))))
  
(defun expr-simplify-1 (expr)
  (cond-matchg
   expr
   ((('+ &rest ?args))
    (expr-simplify-plus-distr
     (expr-simplify-assoc-com '+ ?args 0)))

   ((('* &rest ?args))
    (expr-simplify-assoc-com '* ?args 1))

   ((('- &rest ?args))
    (cond
      ((single ?args) (expr-negate (car ?args)))
      (t
       (expr-simplify-1
	(cons '+ (cons (car ?args) (mapcar #'expr-negate (cdr ?args))))))))

   ((('/= ?c 0) (expr-is-condition ?c))
    (expr-simplify-cond ?c))

   ((('= ?c 0) (expr-is-condition ?c))
    (negate-condition (expr-simplify-cond ?c)))


   ((('= ('< ?a 0) ('overflow-sub ?b ?c))
     (equal ?a (expr-simplify-1 (list '- ?b ?c))))
    (expr-simplify-1 (list '>= ?b ?c)))

   ((('/= ('< ?a 0) ('overflow-sub ?b ?c))
     (equal ?a (expr-simplify-1 (list '- ?b ?c))))
    (expr-simplify-1 (list '< ?b ?c)))

   (((?op ('if ?c ?t1 ?e1) ('if ?c ?t2 ?e2)))
    (expr-simplify-1 (list 'if ?c
			   (expr-simplify-1 (list ?op ?t1 ?t2))
			   (expr-simplify-1 (list ?op ?e1 ?e2)))))

   (((?cmp ?x ?y)
     (member ?cmp '(= /= < > <= >= <u >u <=u >=u)))
    (expr-simplify-cmp expr))

   ((('and &rest ?a))
    (expr-simplify-and-or ?a t))
   ((('or &rest ?a))
    (expr-simplify-and-or ?a nil))

   ((('if 1 ?t ?e)) ?t)
   ((('if 0 ?t ?e)) ?e)

;;    ((('if ('if ?c ?tc ?ec) ?t ?e))
;;     (expr-simplify-1
;;      (list 'if (expr-simplify-1
;; 		(list 'or
;; 		      (expr-simplify-1 (list 'and ?c ?tc))
;; 		      (expr-simplify-1 (list 'and (negate-condition ?c) ?ec))))
;; 	   ?t ?e)))

;;    ;; TODO: is it OK for all ?a ?b, or are some
;;    ;; additional conditions required?
;;    ((('and ('>= ?a ?b) ('< ?c ?d))
;;      (equal ?c (expr-simplify-1 (list '- ?a ?b))))
;;     (list '< ?a (expr-simplify-1 (list '+ ?b ?d))))

;;    ;; TODO: is it OK for all ?a ?b, or are some
;;    ;; additional conditions required?
;;    ((('and ('>=u ?a ?b) ('>=u ?c ?d))
;;      (equal ?c (expr-simplify-1 (list '- ?a ?b))))
;;     (list '>=u ?a (expr-simplify-1 (list '+ ?b ?d))))

;;    ((('or ('< ?a ?c1) ('< ?a ?c2))
;;      (and (numberp ?c1) (numberp ?c2)))
;;     (list '< ?a (max ?c1 ?c2)))

;;    ((('or (?cmp ?a ?b) ('= ?c ?d))
;;      (and (member ?cmp '(< <u))
;; 	  (let ((e1 (expr-simplify-1 (list '- ?a ?b)))
;; 		(e2 (expr-simplify-1 (list '- ?c ?d))))
;; 	    (or (equal e1 e2) (equal e1 (expr-negate e2))))))
;;     (list (case ?cmp (< '<=) (<u '<=u)) ?a ?b))

;;    ((('and ('/= ?x ?y) ('>= ?x ?y)))
;;     (list '> ?x ?y))

;;    ((('and ('>= ?a ?b) ('/= ?c ?d))
;;      (let ((ege (expr-simplify-1 (list '- ?a ?b)))
;; 	   (ene (expr-simplify-1 (list '- ?c ?d))))
;;        (or (equal ege ene) (equal ege (expr-negate ene)))))
;;     (list '> ?a ?b))

;;    ((('and (?cmp ?a ?b) ('= ?b ?a))
;;      (member ?cmp '(< > <u >u)))
;;     0)

   ((('overflow-sub ?x 0))
    0)

   (((?sh ?x 0) (member ?sh '(lsl lsr asr ror)))
    ?x)

   ((('lsl ?x ?n) (numberp ?n))
    (expr-simplify-1 (list '* ?x (ash 1 ?n))))

   ((('lsr ('* ?n &rest ?l) ?sr)
     (and (numberp ?n) (power2p ?n) (numberp ?sr)))
    (let ((sl (power2p ?n)))
      (cond
	((= sl ?sr)
	 (list 'logand (1- (ash 1 (- 32 sl)))
	       (if (single ?l) (car ?l) (cons '* ?l))))
	((< sl ?sr)
	 (list 'lsr
	       (list 'logand (1- (ash 1 (- 32 sl)))
		     (if (single ?l) (car ?l) (cons '* ?l)))
	       (- ?sr sl)))
	(t
	 (list 'logand (1- (ash 1 (- 32 ?sr)))
	       (cons '* (cons (* ?n (ash 1 (- sl ?sr))) ?l)))))))
	      
   ((('if ?c ?t ?e))
    (let ((c (expr-simplify-cond ?c))
	  (then (expr-simplify-with-true (list ?c) ?t))
	  (else (expr-simplify-with-true (list (negate-condition ?c)) ?e)))
      (if (and (eq ?c c) (eq ?t then) (eq ?e else))
	  expr
	  (expr-simplify-1 (list 'if c then else)))))

   ((_) expr)))

(defun expr-simplify (expr)
  (cond
    ((consp expr)
     (expr-simplify-1 (mapcar-reuse #'expr-simplify expr)))
    (t expr)))

(defun find-subexpr (pred expr &key ignore-top)
  (cond
    ((and (not ignore-top) (funcall pred expr))
     expr)
    ((consp expr)
     (let ((r (find-subexpr pred (car expr))))
       (if r r (find-subexpr pred (cdr expr) :ignore-top t))))))

(defun find-all-subexpr (pred expr &key ignore-top)
  (let ((res
	 (if (and (not ignore-top) (funcall pred expr))
	     (list expr)
	     nil)))
    (when (consp expr)
      (append
       (find-all-subexpr pred (car expr))
       (find-all-subexpr pred (cdr expr) :ignore-top t)
       res))))

;; (defun state-remove-unknown (state)
;;   (remove-if
;;    #'(lambda (x) (find-subexpr
;; 		  #'(lambda (y)
;; 		      (or
;; 		       (and (consp y) (eq (car y) 'reg))
;; 		       (eq y 'unknown)))
;; 		  (cdr x)))
;;    state))

(defun atom-or-vref (x)
  (or (atom x) (eq (car x) 'vref)))

(defun expr-substitute-state (expr state)
  (expr-substitute expr state
		   #'atom-or-vref))

(defun frame-expr-simplify (expr)
  (aexpr-rec expr
	     (cond-matchg it
			  ((('mem ('+ ?n ('val 'frame))) (numberp ?n))
			   (list 'frame ?n))
			  ((('mem ('val 'frame)))
			   (list 'frame 0))
			  ((_) (rec-all)))))

(defun state-substitute-simplify (state sub)
  (labels ((simpsub (x)
	     (expr-simplify
	      (expr-substitute x sub #'atom-or-vref))))
    (mapcar #'(lambda (x)
		(cons
		 (if (consp (car x))
		     (cons (caar x)
			   (mapcar #'simpsub (cdar x)))
		     (car x))
		 (simpsub (cdr x))))
	    state)))

(defun sem-create-frame-refs (sem)
  (mapcar #'frame-expr-simplify
	  sem))

(defun stmt-simplify (stmt &optional true)
  (case (car stmt)
    (set (if true
	     (list 'set (second stmt)
		   (expr-simplify-with-true true (third stmt)))
	     stmt))
    (if-then 
     (let ((c (expr-simplify-cond (second stmt))))
       (list 'if-then c
	     (stmt-simplify (third stmt) (cons c true)))))
    (return (if true
		(list 'return
		      (expr-simplify-with-true true (second stmt)))
		stmt))
    ((call callret)
     (if true
	 (cons (car stmt)
	       (expr-simplify-with-true true (cdr stmt)))
	 stmt))
    (t stmt)))

(defun stmt-substitute-state (stmt state &key max-depth)
  (labels ((subst-simplify (expr)
	     (expr-subst-simplify expr state
				  :atom? #'atom-or-vref
				  :max-depth max-depth)))
    (case (car stmt)
      (set (list 'set
		 (cons (car (second stmt))
		       (subst-simplify (cdr (second stmt))))
		 (subst-simplify (third stmt))))
      (if-then (list 'if-then
		     (subst-simplify (second stmt))
		     (stmt-substitute-state (third stmt) state
					    :max-depth max-depth)))
      (t (subst-simplify stmt)))))
  
(defun sem-substitute-state (sem state &key max-depth)
  (mapcar #'(lambda (s)
	      (stmt-simplify
	       (stmt-substitute-state s state :max-depth max-depth)))
	  sem))

(defun sem-substitute (sem sub)
  (mapcar #'(lambda (x)
	      (cons (car x)
		    (expr-substitute (cdr x) sub)))
	  sem))

(defun expr-depends-on-assigned (expr sem)
  (find-subexpr
   #'(lambda (x)
       (find-if
	#'(lambda (y) (equal (car y) x))
	sem))
   expr))

(defun stored-in-state (expr)
  (and (consp expr) (member (car expr) '(reg frame))))

(defun expr-complete-cond-assign (expr dest state pc)
  (if-match ('if ?c ?t) expr
	    (expr-simplify-1
	     (list 'if ?c ?t
		   (or (state-get state dest)
		       (list 'vref dest pc))))
	    expr))

(defun sem-complete-cond-assign (sem state pc)
  (mapcar #'(lambda (p)
	      (cons (car p)
		    (expr-complete-cond-assign (cdr p) (car p)
					       state pc)))
	  sem))

(defun modify-state-by-sem (state sem)
  (labels ((assignments (stmt &optional unknown)
	     (case (car stmt)
	       (set (if (stored-in-state (second stmt))
			(list (cons (second stmt)
				    (if unknown
					'unknown
					(third stmt))))
			()))
	       (if-then (assignments (third stmt) t))
	       (t ()))))
    (let ((st1 (mapcan #'assignments sem)))
      (remove 'unknown 
	      (append st1
		      (set-difference state st1
				      :test #'equal :key #'car))
	      :key #'cdr))))

(defun state-set-reg (state reg value)
  (modify-state-by-sem state
		       (list (cons reg value))))

(defun expr-substitute-rom (expr image)
  (cond
    ((consp expr)
     (let ((expr2
	    (expr-simplify-1
	     (cons (car expr)
		   (mapcar #'(lambda (x)
			       (expr-substitute-rom x image)) (cdr expr))))))
       (cond
	 ((and (consp expr2) (eq (car expr2) 'mem))
	  (if (and (numberp (cadr expr2))
		   (image-defined-word image (cadr expr2)))
	      (image-get-word image (cadr expr2))
	      expr2))
	 (t expr2))))
    (t expr)))

(defun stmt-substitute-rom (stmt image)
  (case (car stmt)
    (set
     (list 'set (second stmt)
	   (expr-substitute-rom (third stmt) image)))
    (if-then
     (list 'if-then (expr-substitute-rom (second stmt) image)
	   (stmt-substitute-rom (third stmt) image)))
    (t (expr-substitute-rom stmt image))))

					     
(defun sem-substitute-rom (sem image)
  (mapcar #'(lambda (x) (stmt-substitute-rom x image))
	  sem))

;; (defun make-var-generator ()
;;   (let ((n 0))
;;     #'(lambda () (list 'var (setf n (1+ n))))))

;; (defun sem-should-be-var (expr)
;;   (or
;;    (eq expr 'unknown)
;;    (and (consp expr)
;; 	(eq (car expr) 'mem))))

;; (defun sem-create-one-var (expr vargen &key ignore-top)
;;   (cond
;;    ((and (not ignore-top) (sem-should-be-var expr))
;;     (let ((v (funcall vargen)))
;;       (cons (cons expr v) v)))
;;    ((consp expr)
;;     (let ((r (sem-create-one-var (car expr) vargen)))
;;       (if r
;; 	  (cons (car r) (cons (cdr r) (cdr expr)))
;; 	(let ((r (sem-create-one-var (cdr expr) vargen :ignore-top t)))
;; 	  (if r
;; 	      (cons (car r) (cons (car expr) (cdr r))))))))))

;; (defun sem-create-vars (sem vargen)
;;   (labels ((rec (sem acc)
;; 		(if (null sem)
;; 		    acc
;; 		  (let ((p (sem-create-one-var (cdar sem) vargen)))
;; 		    (if (null p)
;; 			(rec (cdr sem) (cons (car sem) acc))
;; 		      (if (eq (caar p) 'unknown)
;; 			  (rec (cons (cons (caar sem) (cdr p))
;; 				     (cdr sem))
;; 			       (cons (cons (cdar p) (caar p)) acc))
;; 			(rec
;; 			 (sem-substitute
;; 			  (cons (cons (caar sem) (cdr p)) (cdr sem))
;; 			  (list (car p)))
;; 			 (cons (cons (cdar p) (caar p)) acc))))))))
;;     (rec sem nil)))

;; (defun state-clean (state)
;;   (remove-if
;;    #'(lambda (x) (or (not (numberp (cdr x)))
;; 		     (equal '(reg 15) (car x))
;; 		     (not (eq 'reg (caar x)))))
;;    state))

;; (defun print-state-with-consts (state &optional (stream t))
;;   (do ((l state (cdr l)))
;;       ((not l))
;;       (format stream "r~d=0x~x" (cadaar l) (cdar l))
;;       (if (cdr l) (format stream ","))))

;; (defun find-memory-ref-handler (pc instr sem state1 state2)
;;   (dolist (p sem)
;;     (dolist (m (find-all-subexpr
;; 		#'(lambda (x) (and (consp x) (eq (car x) 'mem)))
;; 		(cdr p)))
;;       (if (numberp (second m))
;; 	  (format t "~8x: R ~8x~%" pc (second m))))
;;     (if (and (eq (caar p) 'mem) (numberp (cadar p)))
;; 	(format t "~8x: W ~8x~%" pc (cadar p)))))

;; (defun verbose-handler (pc instr sem state1 state2)
;;   (format t "~8x: i:~a sm:~a st:~a~%" pc (car instr) sem state1))

;; (defun analyse-range (image start end state handler)
;;   (do ((vargen (make-var-generator)))
;;       ((> start end))
;;       (let* ((i (arm-disassemble-instruction
;; 		 (image-get-sequence image start 4)))
;; 	     (isem
;; 	      (sem-create-vars
;; 	       (sem-substitute-rom
;; 		(sem-substitute-simplify
;; 		 (arm-sem-instruction (car i))
;; 		 (state-set-reg state '(reg 15) (+ start 8)))
;; 		image)
;; 	       vargen))
;; 	     (state2
;; 	      (modify-state-by-sem state isem)))
;; 	(funcall handler start i isem state state2)
;; 	(setf state state2)
;; 	(incf start (cdr i)))))

;; (defun find-memory-refs (image start end &optional state)
;;   (analyse-range image start end state #'find-memory-ref-handler))

;; (defun verbose-analyse (image start end &optional state)
;;   (analyse-range image start end state #'verbose-handler))
  

;;; ---------------- expression printing ---------------------

(defun expr-reg-to-string (expr)
  (cond ((numberp (second expr))
	 (case (second expr)
	   (14 "LR")
	   (15 "PC")
	   (t (format nil "R~d" (second expr)))))
	((symbolp (second expr))
	 (symbol-name (second expr)))
	(t "R?")))

(defun expr-to-string-infix (op args prio &optional (this-prio prio))
  (labels ((rec (res first args)
	     (if (null args)
		 res
		 (if first
		     (rec (expr-to-string (car args) prio) nil (cdr args))
		     (rec
		      (concatenate 'string
				   res
				   op
				   (expr-to-string (car args) prio))
		      nil (cdr args))))))
    (cons (rec "" t args) this-prio)))

(defun strcat (&rest rest)
  (apply #'concatenate (cons 'string rest)))

(defun expr-to-string (expr &optional (prio 1000))
  (let ((res
	 (cond
	   ((numberp expr)
	    (cons
	     (cond 
	       ((< expr 0) (strcat "-" (expr-to-string (- expr))))
	       ((< expr 10) (format nil "~d" expr))
	       (t (format nil "$~x" expr)))
	     0))

	   ((symbolp expr)
	    (cons (symbol-name expr) 0))

	   ((consp expr)
	    (case (car expr)
	      (reg
	       (cons (expr-reg-to-string expr) 0))

	      ((mem memb memh memsb memsh)
	       (cons
		(strcat "[" (expr-to-string (second expr))
			(case (car expr)
			  (mem "")
			  (memb ".B")
			  (memh ".H")
			  (memsb ".SB")
			  (memsh ".SH"))
			"]")
		0))
	     
	      (vref
	       (cons
		(strcat
		 (expr-to-string (second expr))
		 (format nil "@#x~x" (third expr)))
		500))

	      (+
	       (expr-to-string-infix " + " (cdr expr) 300))
	     
	      (*
	       (expr-to-string-infix " * " (cdr expr) 297))
	     
	      (-
	       (if (single (cdr expr))
		   (cons (format nil "-~a" (expr-to-string (cadr expr) 300))
			 300)
		   (expr-to-string-infix " - " (cdr expr) 300)))
	     
	      ((= /= < > <= >= <u <=u >u >=u)
	       (expr-to-string-infix
		(strcat " " (symbol-name (car expr)) " ")
		(cdr expr) 600))
	     
	      (logand
	       (expr-to-string-infix " & " (cdr expr) 250))
	     
	      (logior
	       (expr-to-string-infix " | " (cdr expr) 275))
	     
	      (logxor
	       (expr-to-string-infix " ^ " (cdr expr) 275))
	     
	      (lognot
	       (cons
		(strcat "~" (expr-to-string (second expr) 200))
		200))
	      

	      ((lsl lsr asr ror)
	       (expr-to-string-infix
		(strcat " " (symbol-name (car expr)) " ")
		(cdr expr) 295 1000))

	      ((if)
	       (cons
		(strcat
		 "if " (expr-to-string (second expr) 900)
		 " then " (expr-to-string (third expr) 900)
		 (if (fourth expr)
		     (strcat " else "  (expr-to-string (fourth expr) 900))
		     ""))
		900))

	      (t (cons
		  (strcat (symbol-name (car expr)) "("
			  (car (expr-to-string-infix ", " (cdr expr) 1000))
			  ")")
		  0))))
	   (t (cons (format nil "?~a" expr)  0)))))
    (if (<= prio (cdr res))
	(format nil "(~a)" (car res))
	(car res))))

(defun expr-find-hints (expr state &optional (simplify #'expr-simplify))
  (let ((simp (funcall simplify expr)))
    (cond
      ((consp simp)
       (let ((e1
	      (mapcar #'(lambda (x)
			  (funcall simplify
				   (expr-substitute-state x state)))
		      simp)))
	 (let ((e2 (funcall simplify (expr-substitute-state e1 state))))
	   (if (equal e1 e2)
	       (reduce #'append simp
		       :initial-value nil
		       :key #'(lambda (x) (expr-find-hints x state)))
	       (list (cons simp e2)))))))))

(defun sem-find-hints (sem state &optional (simplify #'expr-simplify))
  (let ((my-union #'(lambda (x y) (union x y :key #'car :test #'equal))))
    (reduce my-union sem
	    :initial-value nil
	    :key #'(lambda (x)
		     (append
		      (when (consp (car x))
			(reduce #'append (car x)
				:initial-value nil
				:key #'(lambda (x)
					 (expr-find-hints x state simplify))))
		      (expr-find-hints (cdr x) state simplify))))))

;;; ---------------- graph analysis -------------------

(defun make-graph ()
  (cons (make-hash-table) (make-hash-table)))

(defun graph-incoming-edges (graph v)
  (gethash v (cdr graph)))

(defun graph-outgoing-edges (graph v)
  (gethash v (car graph)))

(defun graph-add-edge (graph a b data)
  (let ((e (cons (cons a b) data)))
    (setf (gethash a (car graph))
	  (cons e (gethash a (car graph))))
    (setf (gethash b (cdr graph))
	  (cons e (gethash b (cdr graph))))))

(defun graph-remove-edge (graph e)
  (setf (gethash (caar e) (car graph))
	(remove e (gethash (caar e) (car graph))))
  (setf (gethash (cdar e) (cdr graph))
	(remove e (gethash (cdar e) (cdr graph)))))

(defun state-get (state elt)
  (let ((p (assoc elt state :test #'equal)))
    (if p (cdr p))))

(defun sem-get (sem elt)
  (state-get sem elt))

(defun state-intersection (st1 st2)
  (intersection st1 st2 :test #'equal))

(defun sem-incoming-state (graph address)
  (let ((inl (graph-incoming-edges graph address)))
    (when inl
      (labels ((rec (inl state)
		 (if (null inl)
		     state
		     (rec (cdr inl)
			  (state-intersection state (cdr (car inl)))))))
	(rec (cdr inl) (cdr (car inl)))))))

(defun sem-image-instruction-sem (image address)
  (arm-sem-instruction
   (car
    (arm-disassemble-instruction (image-get-sequence image address 4)))))

(defun sem-memory-or-reg (expr)
  (and
   (consp expr)
   (member (car expr) '(reg mem))))

(defun assoc-get (item alist &optional default)
  (let ((a (assoc item alist)))
    (if a (cdr a) default)))

(defun call-args-list (n)
  (labels ((rec (res n)
	     (if (<= n 0)
		 res
		 (rec (cons (if (<= n 4)
				(list 'reg (1- n))
				(list 'mem (list '+ (* 4 (- n 5))
						 (list 'reg 13))))
			    res)
		      (1- n)))))
    (rec () n)))

(defun call-statement (dest pc subroutines)
  (let ((subroutine (assoc-get dest subroutines)))
    (nconc
     (list 'call
	   (assoc-get 'name subroutine dest))
     (let ((args (assoc-get 'args subroutine
			    '(a b c d e))))
       (call-args-list (length args))))))

(defun stmt-dest (stmt)
  (cond-match stmt
	      (('set ?d _) ?d)
	      (('if-then _ ?s) (stmt-dest ?s))
	      (_ nil)))

(defun sem-get-reg-stmt (sem reg)
  (labels ((reg-stmt-p (stmt)
	     (case (car stmt)
	       (set (equal (cadr stmt) reg))
	       (if-then (reg-stmt-p (caddr stmt))))))
    (find-if #'reg-stmt-p sem)))

(defun sem-get-set-expr (sem dest)
  (let ((stmt (sem-get-reg-stmt sem dest)))
    (if-match ('set ?d ?x) stmt
	      ?x
	      nil)))

(defun arm-sem-call-dest-cond (sem state pc inside-p)
  (let ((next-pc (+ pc 4))
	(sem-pc (sem-get-reg-stmt sem '(reg 15)))
	(sem-lr (sem-get-reg-stmt sem '(reg 14))))
    (let ((dc (cond-matchg
	       sem-pc
	       ((('set _ ('if ?c ?t _)))
		(if (and (numberp ?t) (funcall inside-p ?t))
		    nil
		    (cons ?t ?c)))
	       ((('set _ ?x)
		 (not (and (numberp ?t) (funcall inside-p ?x))))
		(cons ?x t))
	       ((_) nil))))
      (when dc
	(cond-matchg
	 sem-lr
	 ((nil)
	  (let ((lr (state-get state '(reg 14))))
	    (if (and lr
		     (eql (expr-simplify-with-true (list (cdr dc)) lr)
			  next-pc))
		dc
		nil)))
	 ((('if-then ?c ('set _ ?x)))
	  (if (and (equal (cdr dc) ?c)
		   (numberp ?x)
		   (= ?x next-pc))
	      dc
	      nil))
	 ((('set _ ?x))
	  (if (and (numberp ?x) (= ?x next-pc))
	      dc
	      nil))
	 ((_) nil))))))

(defun sem-simple-call-semantics (dest pc subroutines scratch)
  (cons
   (call-statement dest pc subroutines)
   (mapcar #'(lambda (x) (list 'set x 'unknown)) scratch)))

(defun sem-arm-call-semantics (cond dest pc subroutines)
  (let ((simple-sem
	 (sem-simple-call-semantics
	  dest
	  pc
	  subroutines
	  '((reg 0) (reg 1) (reg 2) (reg 3) (reg 14)
	    (reg n) (reg z) (reg v) (reg c)))))
    (if (eq cond t)
	simple-sem
	(mapcar #'(lambda (s)
		    (list 'if-then cond s))
		simple-sem))))

(defun sem-set-call-semantics (sem state pc subroutines inside-p)
  (let ((dc (arm-sem-call-dest-cond sem state pc
				    (if inside-p
					inside-p
					#'(lambda (a)
					    (and (>= a pc)
						 (<= a (+ pc 4))))))))
    (if (null dc)
	sem
	(let ((call-sem
	       (sem-arm-call-semantics (cdr dc) (car dc)
				       pc subroutines)))
	  (cons (list 'set (list 'reg 15) (+ pc 4))
		call-sem)))))

(defun sem-set-return-sem (sem pc)
  (let ((next-pc (+ pc 4))
	(sem-pc (sem-get-reg-stmt sem '(reg 15))))
    (cond-match
     sem-pc
     (('set _ ('if ?c ('val 'return) next-pc))
      (cons (list 'set (list 'sreg 'retval) (list 'reg 0))
	    sem))
     (('set _ ('val 'return))
      (cons (list 'set (list 'sreg 'retval) (list 'reg 0))
	    sem))
     (_ sem))))

(defun sem-set-call-ret-sem (sem pc subroutines inside-p)
  (if (null inside-p)
      sem
      (let ((sem-pc (sem-get-reg-stmt sem '(reg 15)))
	    (next-pc (+ pc 4)))
	(cond-matchg
	 sem-pc
	 ((('set _ ('if ?c ?d next-pc))
	   (and (numberp ?d) (not (funcall inside-p ?d))))
	  (nconc
	   (list (list 'set (list 'reg 15)
		       (list 'if ?c (list 'reg 14) next-pc))
		 (list 'set (list 'sreg 'retval) (list 'sreg 'call-result)))
	   (sem-arm-call-semantics ?c ?d pc subroutines)))
	 ((('set _ ?d) (and (numberp ?d) (not (funcall inside-p ?d))))
	  (nconc
	   (list (list 'set (list 'reg 15) (list 'reg 14))
		 (list 'set (list 'sreg 'retval) (list 'sreg 'call-result)))
	   (sem-arm-call-semantics t ?d pc subroutines)))
	 ((_) sem)))))
      
(defun sem-convert-jumps (sem state pc subroutines &key inside-p)
  (sem-set-call-ret-sem
   (sem-set-return-sem
    (sem-set-call-semantics sem state pc subroutines inside-p)
    pc)
   pc
   subroutines
   inside-p))

  

(defun expr-create-vrefs (expr pc &key ignore-top)
  (cond
    ((and (not ignore-top) (stored-in-state expr))
     (list 'vref expr pc))
    ((consp expr)
     (cond
       ((and (not ignore-top) (eq (car expr) 'vref))
	expr)
       (t
	(cons (expr-create-vrefs (car expr) pc)
	      (expr-create-vrefs (cdr expr) pc :ignore-top t)))))
    (t expr)))

(defun stmt-create-vrefs (stmt pc)
  (case (car stmt)
    (set (list 'set
	       (expr-create-vrefs (second stmt) pc
				  :ignore-top t)
	       (expr-create-vrefs (third stmt) pc)))
    (if-then (list 'if-then
		   (expr-create-vrefs (second stmt) pc)
		   (stmt-create-vrefs (third stmt) pc)))
    (call (cons 'call (expr-create-vrefs (cdr stmt) pc)))))

(defun sem-create-vrefs (sem pc)
  (mapcar #'(lambda (s) (stmt-create-vrefs s pc))
	  sem))

(defun stmt-remove-if-then (stmt state pc)
  (cond-matchg stmt
	       ((('if-then ?c ('set ?d ?t)))
		(list 'set ?d
		      (expr-simplify-1
		       (list 'if ?c ?t
			     (let ((v (state-get state ?d)))
			       (if v v
				   (list 'vref ?d pc)))))))
	       ((?x) ?x)))

(defun sem-remove-if-then (sem state pc)
  (mapcar #'(lambda (s)
	      (stmt-remove-if-then s state pc))
	  sem))

(defun state-simplify-cond (true state)
  (mapcar #'(lambda (x) (cons (car x) (expr-simplify-with-true true (cdr x))))
	  state))

(defun seq (from to)
  (if (> from to)
      nil
      (cons from (seq (1+ from) to))))

(defun possible-values-under-cond (expr true)
  (let ((c (some #'(lambda (x)
		     (if-match ('<=u expr ?n) x
			       (if (numberp ?n) ?n)))
		 true)))
    (if c (seq 0 c)
	'unknown)))

(defun depth (x)
  (cond
    ((consp x)
     (1+ (reduce #'max x :key #'depth)))
    (t 1)))

(defun sem-remove-too-deep (sem limit)
  (if limit
      (mapcar #'(lambda (x) (if (> (depth (cdr x)) limit)
				(cons (car x) 'unknown)
				x))
	      sem))
  sem)

(defun state-remove-vrefs-to-pc (state pc)
  (remove-if #'(lambda (p)
		 (find-subexpr #'(lambda (x)
				   (and (consp x)
					(eq (car x) 'vref)
					(eql (caddr x) pc)))
			       (cdr p)))
	     state))

(defun sem-instruction-outgoing-edges (sem state)
  (let* ((pc (sem-get-set-expr sem '(reg 15)))
	 (newst (modify-state-by-sem state sem)))
    (when pc
      (labels ((rec (expr state true)
		 (let ((vl (possible-values-under-cond expr true)))
		   (cond
		     ((not (eq vl 'unknown))
		      (mapcar #'(lambda (x) (cons x state)) vl))
		     ((numberp expr) (list (cons expr state)))
		     ((consp expr)
		      (cond-matchg expr
				   ((('if ?c ?t ?e))
				    (append
				     (rec ?t
					  (state-simplify-cond (list ?c)
							       state)
					  (cons ?c true))
				     (let ((neg (negate-condition ?c)))
				       (rec ?e
					    (state-simplify-cond (list neg)
								 state)
					    (cons neg true)))))
				   (((?op ?n ?x) (and (numberp ?n)
						      (member ?op '(+ *))))
				    (let ((r (rec ?x state true)))
				      (mapcar #'(lambda (x)
						  (cons (funcall
							 (symbol-function ?op)
							 (car x) ?n)
							(cdr x)))
					      r)))))))))
	(rec pc newst nil)))))

(defun full-instruction-sem (image pc state &key subroutines inside-p)
  (let ((st (modify-state-by-sem state
				 (list (list 'set '(reg 15) (+ pc 8))))))
    (sem-create-vrefs
     (sem-substitute-state
      (sem-create-frame-refs
       (sem-substitute-state
	(sem-convert-jumps
	 (sem-substitute-state
	  (sem-create-frame-refs
	   (sem-substitute-rom
	    (sem-substitute-state
	     (sem-image-instruction-sem image pc)
	     st)
	    image))
	  st)
	 st
	 pc
	 subroutines
	 :inside-p inside-p)
	st :max-depth 3))
      st :max-depth 1)
     pc)))

(defun print-semantics (image start end &key subroutines)
  (do ((adr start (+ 4 adr)))
      ((>= adr end))
    (format t "~x: ~a~%" adr (full-instruction-sem image adr ()
						   :subroutines subroutines))))

(defun sem-remove-not-stored (sem)
  (mapcar #'(lambda (s)
	      (if-match ('set ?d ?x) s
			(if (find-subexpr #'(lambda (x)
					      (and (consp x)
						   (or (member (car x)
							       '(mem memb memh memsh memsb)))))
					  ?x)
			    (list 'set ?d 'unknown)
			    s)
			s))
	  sem))
		 
(defun state-add-vrefs (state isem pc)
  (dolist (p isem)
    (dolist (expr (list (car p) (cdr p)))
      (dolist (vr (find-all-subexpr #'(lambda (x)
					(and (consp x)
					     (eq (car x) 'vref)
					     (eql (third x) pc)))
				    expr))
	(unless (state-get state (second vr))
	  (push (cons (second vr) vr) state)))))
  state)

(defun nil-or-vref-to-itself (pc reg expr)
  (or (null expr)
      (and (consp expr)
	   (eq (car expr) 'vref)
	   (equal (cadr expr) reg)
	   (eql (caddr expr) pc))))

(defun analyse-visit-instruction (semf graph states address
				  &key max-depth)
					;(format t "~x~%" address)
  (let* ((current (gethash address states 'undefined))
	 (state (sem-incoming-state graph address))
	 (sem1 (funcall semf address state))
	 (sem2 
	  (sem-remove-not-stored
	   (sem-remove-if-then sem1 state address)))
	 (statevr (sort (copy-list
			 (state-add-vrefs
			  (if (eq current 'undefined)
			      state
			      (remove-if
			       #'(lambda (p)
				   (nil-or-vref-to-itself
				    address
				    (car p)
				    (state-get current (car p))))
			       state))
			  sem2 address))
			#'reg< :key #'car))
	 (isem (sem-remove-too-deep sem2 max-depth)))

    (when (not (equal statevr current))
      (setf (gethash address states) statevr)
      (let ((outl (graph-outgoing-edges graph address))
	    tovisit)
	(dolist (e outl)
	  (graph-remove-edge graph e)
	  (pushnew (cdar e) tovisit))
	(let ((newout
	       (sem-instruction-outgoing-edges isem statevr)))
	  (dolist (e newout)
	    (graph-add-edge graph address (car e) (cdr e))
	    (pushnew (car e) tovisit))
	  tovisit)))))

(defun make-inside-p-min-max (min max)
  #'(lambda (a) (and (or (null min) (<= min a))
		     (or (null max) (> max a)))))

(defun do-analyse (semf address &key state inside-p
		   max-depth)
  (let ((graph (make-graph))
	(states (make-hash-table)))
    (graph-add-edge graph -1 address state)
    (do ((q (list address)))
	((null q) (values states graph))

      (let ((tovisit
	     (if (or (null inside-p) (funcall inside-p (car q)))
		 (analyse-visit-instruction semf
					    graph states (car q)
					    :max-depth max-depth))))
	(setf q (append (cdr q) tovisit))))))

(defun analyse (image address &key state subroutines
		inside-p max-depth)
  (do-analyse #'(lambda (address state)
		  (full-instruction-sem image
					address state
					:subroutines subroutines
					:inside-p inside-p))
    address
    :state state
    :inside-p inside-p
    :max-depth max-depth))

(defun expr-find-vrefs (expr)
  (find-all-subexpr #'(lambda (x) (and (consp x) (eq (car x) 'vref)))
		    expr))

(defun reduce-union-equal (list &key (key #'identity))
  (reduce #'(lambda (a b) (union a b :test #'equal))
	  list
	  :initial-value nil
	  :key key))

(defun exprl-find-vrefs (exprl)
  (reduce-union-equal exprl :key #'expr-find-vrefs))

(defun side-effect-term (expr)
  (and (consp expr)
       (member (car expr) '(mem memb memh memsb memsh call))))

(defun sem-side-effects-used (sem)
  (reduce-union-equal
   sem
   :key #'(lambda (s)
	    (exprl-find-vrefs
	     (find-all-subexpr #'side-effect-term s)))))

(defun stmt-not-stored-set-used (stmt)
  (cond-match
   stmt
   (('set ?d ?x)
    (if (stored-in-state ?d)
	()
	(expr-find-vrefs ?x)))
   (('if-then ?c ('set ?d ?x))
    (if (stored-in-state ?d)
	()
	(append
	 (expr-find-vrefs ?c)
	 (expr-find-vrefs ?x))))
   (_ ())))

(defun sem-not-stored-set-used (sem)
  (reduce-union-equal
   sem
   :key #'stmt-not-stored-set-used))

(defun sem-stored-set-used (sem used pc)
  (reduce-union-equal
   sem
   :key #'(lambda (s)
	    (cond-match s
			(('set ?d ?x)
			 (if (member ?d used :test #'equal)
			     (expr-find-vrefs ?x)
			     ()))
			(('if-then ?c ('set ?d ?x))
			 (if (member ?d used :test #'equal)
			     (append
			      (list (list 'vref ?d pc))
			      (expr-find-vrefs ?x)
			      (expr-find-vrefs ?c))))
			(_ ())))))
			 

(defun sem-pass-through-used (sem state used pc)
  (reduce-union-equal
   used
   :key #'(lambda (r)
	    (if (sem-get-reg-stmt sem r)
		nil
		(list (list 'vref r pc))))))

(defun sem-state-used-registers (sem state usedvals pc)
  (reduce-union-equal
   (list
    (sem-side-effects-used sem)
    (sem-not-stored-set-used sem)
    (sem-stored-set-used sem usedvals pc)
    (sem-pass-through-used sem state usedvals pc))))
	      
(defun add-used-register-at-point (used pc graph reg)
  (let (changed)
    (dolist (e (graph-incoming-edges graph pc))
      (let ((pc1 (caar e)))
					;(format t "adding ~a to used at ~x~%" reg pc1)
	(unless (member reg (gethash pc1 used) :test #'equal)
	  (setf (gethash pc1 used)
		(cons reg (gethash pc1 used)))
	  (push pc1 changed))))
    changed))

(defun sem-is-return (sem)
  (let ((pc (sem-get-set-expr sem '(reg 15))))
    (and pc (find-subexpr #'(lambda (x) (equal x '(val return))) pc))))

(defun mark-instr-used-registers (used image pc state graph subroutines inside-p)
  (let (changed
	(isem (full-instruction-sem image pc state
				    :subroutines subroutines
				    :inside-p inside-p)))
    (let ((usedvalues (sem-state-used-registers isem
						state
						(gethash pc used)
						pc)))
					;(format t "pc: ~x, used: ~a~%" pc (gethash pc used))
      (dolist (v usedvalues)
	(if-match ('vref ?v ?pc) v
		  (setf changed
			(append changed
				(add-used-register-at-point used
							    ?pc graph ?v)))))
      changed)))

(defun mark-used-registers (image states graph subroutines inside-p)
  (let (q
	(used (make-hash-table)))
    (maphash #'(lambda (pc state)
		 (setf (gethash pc used)
		       (list (list 'reg 15)))
		 (push pc q))
	     states)
    (do ()
	((null q) used)
      (let ((changed
	     (mark-instr-used-registers used image (car q)
					(gethash (car q) states)
					graph
					subroutines
					inside-p)))
	(setf q (nconc (cdr q) changed))))))

(defun sem-is-nop (sem used pc)
  (notany #'(lambda (s)
	      (let ((d (stmt-dest s))
		    (next-pc (+ pc 4)))
		(or (find-subexpr #'side-effect-term s)
		    (if-match ('frame _) d t nil)
		    (and (equal d '(reg 15))
			 (if-match ('set _ next-pc) s nil t))
		    (and (not (equal d '(reg 15)))
			 (member d used :test #'equal)))))
	  sem))

(defun mark-used-instructions (image states subroutines usedr)
  (let ((usedi (make-hash-table)))
    (maphash #'(lambda (pc state)
		 (let ((sem (full-instruction-sem
			     image pc state
			     :subroutines subroutines)))
					;(format t "pc: ~x sem: ~a usedr ~a~%" pc sem (gethash pc usedr))
		   (setf (gethash pc usedi)
			 (if (sem-is-nop sem (gethash pc usedr) pc)
			     nil
			     t))))
	     states)
    usedi))

(defun do-print-states (image states handler)
  (let (sl)
    (maphash #'(lambda (key val) (push (cons key val) sl))
	     states)
    (dolist (p (sort sl #'< :key #'car))
      (let* ((pc (car p))
	     (seq (image-get-sequence image pc 4))
	     (instr (arm-disassemble-instruction seq)))
	(funcall handler image pc seq instr (cdr p))))))

(defun default-print-states-f (image pc seq instr state
			       &key subroutines max-address)
  (if (or (null max-address) (< pc max-address))
      (let ((sem
	     (let ((s (arm-sem-instruction (car instr))))
	       (if (member 'call s :key #'car)
		   (sem-set-call-semantics
		    (sem-substitute-simplify s
					     (list
					      (cons '(reg 15)
						    (+ pc 8))))
		    pc
		    subroutines)
		   s)))
	    (byteseq (byteseq-to-string (reverse
					 (subseq seq 0 (cdr instr)))))
	    (dis (arm-instruction-to-string (car instr) pc)))
	(dolist (h (sem-find-hints sem
				   (state-set-reg state '(reg 15)
						  (+ pc 8))
				   #'(lambda (x)
				       (expr-substitute-rom
					(expr-simplify x)
					image))))
	  (if (not (equal '(+ -4 (reg 15)) (car h)))
	      (format t "    ~a == ~a~%"
		      (expr-to-string (car h)) (expr-to-string (cdr h)))))
	(format t "~8x: ~8a ~a~%" pc byteseq dis))))

(defun print-states (image states &key subroutines max-address)
  (do-print-states image states
		   #'(lambda (image pc instr seq state)
		       (default-print-states-f image pc instr seq state
					       :subroutines subroutines :max-address max-address))))

(defun print-entire-states-f (image pc seq instr state)
  (let ((s (arm-sem-instruction (car instr)))
	(byteseq (byteseq-to-string (reverse
				     (subseq seq 0 (cdr instr)))))
	(dis (arm-instruction-to-string (car instr) pc)))
    (dolist (i state)
      (format t "    ~a = ~a~%"
	      (expr-to-string (car i)) (expr-to-string (cdr i))))
    (format t "~8x: ~8a ~a~%" pc byteseq dis)))

(defun print-entire-states (image states)
  (do-print-states image states #'print-entire-states-f))

(defun find-call-in-sem (sem)
  (labels ((rec (sem)
	     (if (null sem)
		 nil
		 (let ((c (find-subexpr #'(lambda (x)
					    (and (consp x)
						 (eq (car x) 'call)))
					(car sem))))
		   (if c (cadr c) (rec (cdr sem)))))))
    (rec sem)))

(defun find-calls-in-states (image states inside-p)
  (let (sub)
    (maphash #'(lambda (pc state)
		 (let ((sem (full-instruction-sem image pc state
						  :inside-p inside-p)))
		   (let ((c (find-call-in-sem sem)))
		     (if (and c (numberp c))
			 (setf sub (adjoin c sub))))))
	     states)
    sub))

(defun called-subroutines (image subroutine subroutines)
  (let ((inside-p (make-inside-p-subroutine subroutines
					    (car subroutine))))
    (find-calls-in-states
     image
     (analyse image (car subroutine)
	      :state (subroutine-initial-state subroutine)
	      :subroutines subroutines
	      :inside-p (make-inside-p-subroutine subroutines
						  (car subroutine))
	      :max-depth (assoc-get 'max-depth (cdr subroutine)))
     inside-p)))

(defun subroutine-by-name (name subroutines)
  (find-if #'(lambda (x) (eq (assoc-get 'name (cdr x)) name))
	   subroutines))

(defun called-new-subroutines (image sub subroutines)
  (format t "looking for subroutines in ~a~%"
	  (assoc-get 'name (cdr sub)))
  (remove-if
   #'(lambda (a) (assoc-get a subroutines))
   (called-subroutines image sub subroutines)))

(defun find-new-subroutines (image subroutines &key file
			     (name "subroutines"))
  (labels ((next (subr)
	     (find-if #'(lambda (s)
			  (assoc-get 'find-sub-new
				     (cdr s)))
		      subr))
	   (default-entry (a)
	       (cons a
		     (list
		      (cons 'name
			    (intern (strcat "P"
					    (format nil "~x" a))))
		      (cons 'args (list 'arg1 'arg2 'arg3 'arg4))
		      (cons 'find-sub-new t))))
	   (sort-subr (subr)
	     (sort (copy-list subr) #'<
		   :key #'car))
	   (save (subr)
	     (let ((f (open file :direction :output
			    :if-exists :supersede
			    :if-does-not-exist :create)))
	       (format f "(defparameter ~a~%'(" name)
	       (format f "~:{(#x~x ~s~@{ ~s~})~%~}" subr)
	       (format f "))~%")
	       (close f)))
	   (rec (subr)
	     (let ((s (next subr)))
	       (if (null s)
		   (if (null file)
		       (sort-subr subr)
		       nil)
		   (let ((nsubr
			  (cons
			   (cons (car s)
				 (remove 'find-sub-new (cdr s)
					 :key #'car))
			   (append (remove s subr :test #'eq)
				   (mapcar #'default-entry
					   (called-new-subroutines
					    image s subr))))))
		     (when file (save (sort-subr nsubr)))
		     (rec nsubr))))))
    (rec subroutines)))

(defun byteseq-to-string (seq)
  (format nil "~{~2,'0x~}" seq))

(defun print-disassemble (image start end)
  (do (step
       (addr start (+ addr step)))
      ((>= addr end))
    (let* ((seq (image-get-sequence image addr 4))
	   (instr (arm-disassemble-instruction seq)))
      (setf step (cdr instr))
      (let ((byteseq (byteseq-to-string (reverse (subseq seq 0 step))))
	    (dis (arm-instruction-to-string (car instr) addr)))
	(format t "~8x: ~8a ~a~%" addr byteseq dis)))))

(defun subroutine-initial-state (sub)
  (labels ((process-args (args n res)
	     (if (null args)
		 res
		 (process-args (cdr args)
			       (1+ n)
			       (cons (cons (if (<= n 4)
					       (list 'reg (1- n))
					       (list 'frame (* 4 (- n 5))))
					   (list 'val (car args)))
				     res)))))
    (process-args (assoc-get 'args (cdr sub)) 1
		  (list
		   (cons '(reg 15) (car sub))
		   (cons '(reg 13) (list 'val 'frame))
		   (cons '(reg 14) (list 'val 'return))))))

(defun print-analysed-subroutine (image sub states subroutines
				  &key max-address)
  (format t ";========================================~%; ~a:~%"
	  (assoc-get 'name (cdr sub) ""))
  (print-states image states :subroutines subroutines
		:max-address max-address))

(defun analyse-subroutines (image subroutines)
  (do ((l (sort (copy-list subroutines) #'< :key #'car) (cdr l)))
      ((null l))
    (unless (or (db-entry-not-analyse (car l))
		(not (db-entry-subroutinep (car l))))
      (let ((states (analyse image (caar l)
			     :state (subroutine-initial-state (car l))
			     :subroutines subroutines
			     :inside-p (make-inside-p-subroutine subroutines
								 (caar l)))))
	(print-analysed-subroutine image (car l) states subroutines
				   :max-address (if (cdr l) (caadr l)))))))

(defun analyse-subroutine (image name subroutines)
  (let ((sub (find-if #'(lambda (x) (eq (assoc-get 'name (cdr x)) name))
		      subroutines)))
    (if sub
	(let ((states (analyse image (car sub)
			       :state (subroutine-initial-state sub)
			       :subroutines subroutines
			       :inside-p (make-inside-p-subroutine subroutines
								   (car sub)))))
	  (print-analysed-subroutine image sub states subroutines)))))

(defun sem-restrict-to-used (sem used)
  (remove-if #'(lambda (s)
		 (let ((d (stmt-dest s)))
		   (and (stored-in-state d)
			(not (member d used :test #'equal))
			(not (find-subexpr #'side-effect-term s))
			(not (match ('frame _) d)))))
	     sem))

(defun propagate-vrefs (image states address usedi usedr subroutines)
  (let ((vrefs (make-hash-table)))
    (maphash #'(lambda (pc state)
		 (dolist (v (expr-find-vrefs state))
		   (let ((r (second v))
			 (pc2 (third v)))
		     (setf (gethash pc2 vrefs)
			   (adjoin r (gethash pc2 vrefs)
				   :test #'equal)))))
	     states)
    (do-analyse #'(lambda (pc state)
		    (let ((sem
			   (if (gethash pc usedi)
			       (sem-restrict-to-used
				(full-instruction-sem image pc
						      (gethash pc states)
						      :subroutines subroutines)
				(gethash pc usedr))
			       (list (list 'set (list 'reg 15) (+ pc 4))))))
		      (dolist (v (gethash pc vrefs))
			(unless (sem-get-reg-stmt sem v)
			  (push (list 'set v (list 'vref v pc)) sem)))
		      sem))
      address
      :inside-p (make-inside-p-subroutine subroutines address))))

(defun semexts-default (sem pc)
  (delete nil
	  (mapcar #'(lambda (p) 
		      (if (equal p (cons '(reg 15) (+ pc 4)))
			  nil
			  (format nil "~a = ~a"
				  (expr-to-string (car p))
				  (expr-to-string (cdr p)))))
		  sem)))

(defun semexts-get-cond (expr pred)
  (cond-matchg expr
	       ((('if ?c ?t _) (funcall pred ?t))
		(cons ?c ?t))
	       ((?x (funcall pred ?x))
		(cons t ?x))
	       ((_) nil)))

(defun semexts-cond (info statement)
  (if (eq (car info) t)
      (list statement)
      (list (strcat "if "
		    (expr-to-string (car info) 900)
		    " then "
		    statement))))

(defun semexts-returnp (sem)
  (semexts-get-cond (sem-get sem '(reg 15))
		    #'(lambda (x)
			(equal x '(val return)))))

(defun semexts-return (sem)
  (semexts-cond (semexts-returnp sem)
		"RETURN R0"))

(defun semexts-callp (sem)
  (semexts-get-cond (sem-get sem '(reg 0))
		    #'(lambda (x)
			(and (consp x)
			     (eq (car x) 'call)))))

(defun semexts-call (sem)
  (let ((p (semexts-callp sem)))
    (semexts-cond p (expr-to-string (cdr p)))))

(defun int-ext-stmt-to-string (stmt)
  (case (car stmt)
    (set (strcat (expr-to-string (second stmt))
		 " := "
		 (expr-to-string (third stmt))))
    (if-then (strcat "IF "
		     (expr-to-string (second stmt))
		     " THEN "
		     (int-ext-stmt-to-string (third stmt))))
    ((call callret) (strcat (symbol-name (car stmt)) "("
			    (car (expr-to-string-infix ", " (cdr stmt) 1000))
			    ")"))
    (return (strcat "RETURN "
		    (expr-to-string (second stmt))))
    (t (format nil "~a" stmt))))

(defun ext-stmt-to-string (stmt)
  (case (car stmt)
    (label (format nil "~a:" (second stmt)))
    (t (strcat "  " (int-ext-stmt-to-string stmt)))))

(defun sem-get-sreg (sem sreg)
  (some #'(lambda (s)
	    (if-match ('set ('sreg sreg) ?x) s ?x nil))
	sem))

(defun ext-sem-make-callret (sem)
  (if-some-match '(return (sreg call-result)) sem
		 (let ((r it))
		   (if-some-match ('call &rest ?args) sem
				  (cons (cons 'callret ?args)
					(set-difference sem (list r it)))
				  sem))
		 (if-some-match
		  ('if-then ?c '(return (sreg call-result))) sem
		  (let ((r it) (c ?c))
		    (if-some-match ('if-then c ('call &rest ?args)) sem
				   (cons (list 'if-then c (cons 'callret ?args))
					 (set-difference sem (list r it)))
				   sem))
		  sem)))

;;   (let ((r (find '(return (sreg call-result)) sem :test #'equal)))
;;     (if r
;; 	(let ((c (find-if #'(lambda (s)
;; 			      (match ('call &rest _) s))
;; 			  sem)))
;; 	  (if c
;; 	      (cons (cons 'callret (cdr c))
;; 		    (set-difference sem (list c r)))
;; 	    sem))
;;       (let ((r (find-if #'(lambda (s)
;; 			    (match ('if _ ('return ('sreg 'call-result))) s)))))
;; 	(if r
;; 	    (let ((c (find-if #'(lambda (s)


(defun expr-put-labels (expr lab)
  (aexpr-rec expr
	     (let ((p (assoc expr lab)))
	       (if p
		   (cdr p)
		   (rec-all)))))

(defun ext-sem-put-labels (sem lab)
  (mapcar #'(lambda (stmt)
	      (cond-match
	       stmt
	       (('set ('reg 15) ?d)
		(list 'set (list 'reg 15)
		      (expr-put-labels ?d lab)))
	       (('if-then ?c ('set ('reg 15) ?d))
		(list 'if-then ?c (list 'set (list 'reg 15)
					(expr-put-labels ?d lab))))
	       (_ stmt)))
	  sem))

(defun sem-to-ext-sem (sem pc lab)
  (labels ((retstmt (sem)
	     (let ((v (sem-get-sreg sem 'retval)))
	       (if v
		   (list 'return v)
		   (list 'return)))))
    (ext-sem-put-labels
     (ext-sem-make-callret
      (mapcan #'(lambda (s) 
		  (let ((next-pc (+ pc 4)))
		    (cond-match
		     s
		     (('set ('reg 15) next-pc) nil)
		     (('set ('sreg _) _) nil)
		     (('set ('reg 15) ('val 'return))
		      (list (retstmt sem)))
		     (('set ('reg 15) ('if ?c ?t next-pc))
		      (list 
		       (stmt-simplify
			(list 'if-then ?c
			      (if-match ('val 'return) ?t
					(retstmt sem)
					(list 'set (list 'reg 15) ?t))))))
		     (_ (list s)))))
	      sem))
     lab)))

(defun expr-put-back-registers (expr state)
  (aexpr-rec expr
	     (if (and (consp it) (not (eq (car it) 'val)))
		 (let ((p (find it state
				:test #'equal
				:key #'cdr)))
		   (if p (car p) (rec-all)))
		 it)))
  

(defun db-entry-not-analyse (entry)
  (assoc-get 'do-not-analyse (cdr entry)))

(defun db-entry-type (entry)
  (assoc-get 'type (cdr entry) 'sub))

(defun db-entry-subroutinep (entry)
  (eq (db-entry-type entry) 'sub))

(defun get-memloc (addr db)
  (assoc addr db))

(defun get-memloc-name (addr db)
  (let ((entry (get-memloc addr db)))
    (if entry
	(assoc-get 'name (cdr entry))
	nil)))

(defun stmt-put-memloc-names (stmt db)
  (aexpr-rec stmt
	     (if (and (consp it)
		      (member (car it) '(mem memb memsb memh memsh))
		      (numberp (cadr it)))
		 (let ((name (get-memloc-name (cadr it) db)))
		   (if name
		       (list (car it) name)
		       (rec-all)))
		 (rec-all))))

(defun sem-put-memloc-names (sem db)
  (mapcar-reuse #'(lambda (x) (stmt-put-memloc-names x db)) sem))

(defun sem-put-back-registers (sem state)
  (mapcar #'(lambda (p)
	      (cons (if (stored-in-state (car p))
			(car p)
			(expr-put-back-registers (car p) state))
		    (expr-put-back-registers (cdr p) state)))
	  sem))

(defun labels-from-states (image states subroutines inside-p)
  (labels ((make-labels (dests n res)
	     (if (null dests)
		 res
		 (make-labels (cdr dests) (1+ n)
			      (cons (cons (car dests)
					  (intern (format nil "L~d" n)))
				    res)))))
    (let (dests)
      (maphash #'(lambda (addr state)
		   (let ((sem (sem-remove-if-then
			       (full-instruction-sem
				image addr state
				:subroutines subroutines
				:inside-p inside-p)
			       state addr)))
		     (setf dests
			   (union dests
				  (remove
				   (+ addr 4)
				   (mapcar #'car
					   (sem-instruction-outgoing-edges
					    sem
					    state)))))))
	       states)
      (make-labels (sort dests #'<) 1 ()))))

(defun ext-sem-adjoin-labels (sem pc1 pc2 lab)
  (nconc (mapcar #'(lambda (x) (list 'label (cdr x)))
		 (sort (remove-if-not #'(lambda (x)
					  (and (> x pc1) (<= x pc2)))
				      lab
				      :key #'car)
		       #'< :key #'car))
	 sem))
	  

(defun print-used-instructions (image states vrefs-states usedi usedr
				subroutines &key inside-p)
  (let ((lab (labels-from-states image states subroutines inside-p)) 
	sl
	(last-pc -1))
    (maphash #'(lambda (key val) (push (cons key val) sl))
	     states)
    (dolist (p (sort sl #'< :key #'car))
      (let ((pc (car p)))
	(when (gethash pc usedi)
	  (let ((dis (ext-sem-adjoin-labels
		      (sem-to-ext-sem
		       (sem-put-memloc-names
			(sem-put-back-registers
			 (sem-restrict-to-used
			  (full-instruction-sem image pc
						(gethash pc states)
						:subroutines subroutines
						:inside-p inside-p)
			  (gethash pc usedr))
			 (gethash pc vrefs-states))
			subroutines)
		       pc lab)
		      last-pc pc lab)))
	    (setf last-pc pc)
	    (format t "~8x: ~a~%" pc (ext-stmt-to-string (car dis)))
	    (dolist (l (cdr dis))
	      (format t "          ~a~%" (ext-stmt-to-string l)))))))))

(defun next-subroutine (subroutines address)
  (labels ((rec (subr res)
	     (if (null subr)
		 res
		 (rec
		  (cdr subr)
		  (if (> (caar subr) address)
		      (if res
			  (min res (caar subr))
			  (caar subr))
		      res)))))
    (rec subroutines nil)))

(defun make-inside-p-subroutine (subroutines address)
  (make-inside-p-min-max address (next-subroutine subroutines address)))

(defun subroutine-ext-syntax (image name subroutines
			      &key max-address)
  (let ((sub (find-if #'(lambda (x) (eq (assoc-get 'name (cdr x)) name))
		      subroutines)))
    (when sub
      (let ((inside-p (make-inside-p-subroutine subroutines (car sub))))
	(multiple-value-bind
	      (states graph) (analyse image (car sub)
				      :state (subroutine-initial-state sub)
				      :subroutines subroutines
				      :inside-p inside-p)
	  (let* ((usedr (mark-used-registers image states
					     graph subroutines inside-p))
		 (usedi (mark-used-instructions image states
						subroutines usedr))
		 (states2 (propagate-vrefs image states (car sub)
					   usedi usedr subroutines)))
	    (print-used-instructions image states states2 usedi usedr
				     subroutines
				     :inside-p inside-p)))))))

(defun subroutines-ext-syntax (image subroutines)
  (do ((l (sort (copy-list subroutines) #'< :key #'car) (cdr l)))
      ((null l))
    (unless (or (db-entry-not-analyse (car l))
		(not (db-entry-subroutinep (car l))))
      (format t ";========================================~%; ~a:~%"
	      (assoc-get 'name (cdar l)))
      (subroutine-ext-syntax image (assoc-get 'name (cdar l))
			     subroutines))))
